# Copyright (c) 2010, 2011, 2012 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import os
import signal
import sys
import time

import ovs.dirs
import ovs.fatal_signal
import ovs.process
import ovs.socket_util
import ovs.timeval
import ovs.util
import ovs.vlog

if sys.platform != 'win32':
    import fcntl
    import resource
else:
    import ovs.winutils as winutils
    import ovs.fcntl_win as fcntl
    import pywintypes
    import subprocess
    import win32process

vlog = ovs.vlog.Vlog("daemon")

# --detach: Should we run in the background?
_detach = False

# Running as the child process - Windows only.
_detached = False

# --pidfile: Name of pidfile (null if none).
_pidfile = None

# Our pidfile's inode and device, if we have created one.
_pidfile_dev = None
_pidfile_ino = None

# --overwrite-pidfile: Create pidfile even if one already exists and is locked?
_overwrite_pidfile = False

# --no-chdir: Should we chdir to "/"?
_chdir = True

# --monitor: Should a supervisory process monitor the daemon and restart it if
# it dies due to an error signal?
_monitor = False

# File descriptor used by daemonize_start() and daemonize_complete().
_daemonize_fd = None

RESTART_EXIT_CODE = 5


def make_pidfile_name(name):
    """Returns the file name that would be used for a pidfile if 'name' were
    provided to set_pidfile()."""
    if name is None or name == "":
        return "%s/%s.pid" % (ovs.dirs.RUNDIR, ovs.util.PROGRAM_NAME)
    else:
        return ovs.util.abs_file_name(ovs.dirs.RUNDIR, name)


def set_pidfile(name):
    """Sets up a following call to daemonize() to create a pidfile named
    'name'.  If 'name' begins with '/', then it is treated as an absolute path.
    Otherwise, it is taken relative to ovs.util.RUNDIR, which is
    $(prefix)/var/run by default.

    If 'name' is null, then ovs.util.PROGRAM_NAME followed by ".pid" is
    used."""
    global _pidfile
    _pidfile = make_pidfile_name(name)


def set_no_chdir():
    """Sets that we do not chdir to "/"."""
    global _chdir
    _chdir = False


def ignore_existing_pidfile():
    """Normally, daemonize() or daemonize_start() will terminate the program
    with a message if a locked pidfile already exists.  If this function is
    called, an existing pidfile will be replaced, with a warning."""
    global _overwrite_pidfile
    _overwrite_pidfile = True


def set_detach():
    """Sets up a following call to daemonize() to detach from the foreground
    session, running this process in the background."""
    global _detach
    _detach = True


def set_detached(wp):
    """Sets up a following call to daemonize() to fork a supervisory
    process to monitor the daemon and restart it if it dies due to
    an error signal. Used on Windows only."""
    global _detached
    global _daemonize_fd
    _detached = True
    _daemonize_fd = int(wp)


def get_detach():
    """Will daemonize() really detach?"""
    return _detach


def set_monitor():
    """Sets up a following call to daemonize() to fork a supervisory process to
    monitor the daemon and restart it if it dies due to an error signal."""
    global _monitor
    _monitor = True


def _fatal(msg):
    vlog.err(msg)
    sys.stderr.write("%s\n" % msg)
    sys.exit(1)


def _make_pidfile():
    """If a pidfile has been configured, creates it and stores the running
    process's pid in it.  Ensures that the pidfile will be deleted when the
    process exits."""
    pid = os.getpid()

    # Create a temporary pidfile.
    if sys.platform != 'win32':
        tmpfile = "%s.tmp%d" % (_pidfile, pid)
        ovs.fatal_signal.add_file_to_unlink(tmpfile)
    else:
        tmpfile = "%s" % _pidfile

    try:
        # This is global to keep Python from garbage-collecting and
        # therefore closing our file after this function exits.  That would
        # unlock the lock for us, and we don't want that.
        global file_handle

        file_handle = open(tmpfile, "w")
    except IOError as e:
        _fatal("%s: create failed (%s)" % (tmpfile, e.strerror))

    try:
        s = os.fstat(file_handle.fileno())
    except IOError as e:
        _fatal("%s: fstat failed (%s)" % (tmpfile, e.strerror))

    try:
        file_handle.write("%s\n" % pid)
        file_handle.flush()
    except OSError as e:
        _fatal("%s: write failed: %s" % (tmpfile, e.strerror))

    try:
        if sys.platform != 'win32':
            fcntl.lockf(file_handle, fcntl.LOCK_EX | fcntl.LOCK_NB)
        else:
            fcntl.lockf(file_handle, fcntl.LOCK_SH | fcntl.LOCK_NB)
    except IOError as e:
        _fatal("%s: fcntl failed: %s" % (tmpfile, e.strerror))

    if sys.platform == 'win32':
        # Ensure that the pidfile will gets closed and deleted on exit.
        ovs.fatal_signal.add_file_to_close_and_unlink(_pidfile, file_handle)
    else:
        # Rename or link it to the correct name.
        if _overwrite_pidfile:
            try:
                os.rename(tmpfile, _pidfile)
            except OSError as e:
                _fatal("failed to rename \"%s\" to \"%s\" (%s)"
                       % (tmpfile, _pidfile, e.strerror))
        else:
            while True:
                try:
                    os.link(tmpfile, _pidfile)
                    error = 0
                except OSError as e:
                    error = e.errno
                if error == errno.EEXIST:
                    _check_already_running()
                elif error != errno.EINTR:
                    break
            if error:
                _fatal("failed to link \"%s\" as \"%s\" (%s)"
                       % (tmpfile, _pidfile, os.strerror(error)))

        # Ensure that the pidfile will get deleted on exit.
        ovs.fatal_signal.add_file_to_unlink(_pidfile)

        # Delete the temporary pidfile if it still exists.
        if not _overwrite_pidfile:
            error = ovs.fatal_signal.unlink_file_now(tmpfile)
            if error:
                _fatal("%s: unlink failed (%s)" % (
                    tmpfile, os.strerror(error)))

    global _pidfile_dev
    global _pidfile_ino
    _pidfile_dev = s.st_dev
    _pidfile_ino = s.st_ino


def daemonize():
    """If configured with set_pidfile() or set_detach(), creates the pid file
    and detaches from the foreground session."""
    daemonize_start()
    daemonize_complete()


def _waitpid(pid, options):
    while True:
        try:
            return os.waitpid(pid, options)
        except OSError as e:
            if e.errno == errno.EINTR:
                pass
            return -e.errno, 0


def _fork_and_wait_for_startup():
    if sys.platform == 'win32':
        return _fork_and_wait_for_startup_windows()

    try:
        rfd, wfd = os.pipe()
    except OSError as e:
        sys.stderr.write("pipe failed: %s\n" % os.strerror(e.errno))
        sys.exit(1)

    try:
        pid = os.fork()
    except OSError as e:
        sys.stderr.write("could not fork: %s\n" % os.strerror(e.errno))
        sys.exit(1)

    if pid > 0:
        # Running in parent process.
        os.close(wfd)
        ovs.fatal_signal.fork()
        while True:
            try:
                s = os.read(rfd, 1)
                error = 0
            except OSError as e:
                s = ""
                error = e.errno
            if error != errno.EINTR:
                break
        if len(s) != 1:
            retval, status = _waitpid(pid, 0)
            if retval == pid:
                if os.WIFEXITED(status) and os.WEXITSTATUS(status):
                    # Child exited with an error.  Convey the same error to
                    # our parent process as a courtesy.
                    sys.exit(os.WEXITSTATUS(status))
                else:
                    sys.stderr.write("fork child failed to signal "
                                     "startup (%s)\n"
                                     % ovs.process.status_msg(status))
            else:
                assert retval < 0
                sys.stderr.write("waitpid failed (%s)\n"
                                 % os.strerror(-retval))
                sys.exit(1)

        os.close(rfd)
    else:
        # Running in parent process.
        os.close(rfd)
        ovs.timeval.postfork()

        global _daemonize_fd
        _daemonize_fd = wfd
    return pid


def _fork_and_wait_for_startup_windows():
    global _detached
    if _detached:
        # Running in child process
        ovs.timeval.postfork()
        return 0

    """ close the log file, on Windows cannot be moved while the parent has
    a reference on it."""
    vlog.close_log_file()

    try:
        (rfd, wfd) = winutils.windows_create_pipe()
    except pywintypes.error as e:
        sys.stderr.write("pipe failed to create: %s\n" % e.strerror)
        sys.exit(1)

    try:
        creationFlags = win32process.DETACHED_PROCESS
        args = ("%s %s --pipe-handle=%ld" % (
            sys.executable, " ".join(sys.argv), int(wfd)))
        proc = subprocess.Popen(
            args=args,
            close_fds=False,
            shell=False,
            creationflags=creationFlags,
            stdout=sys.stdout,
            stderr=sys.stderr)
        pid = proc.pid
    except OSError as e:
        sys.stderr.write("CreateProcess failed (%s)\n" % os.strerror(e.errno))
        sys.exit(1)

    # Running in parent process.
    winutils.win32file.CloseHandle(wfd)
    ovs.fatal_signal.fork()

    error, s = winutils.windows_read_pipe(rfd, 1)
    if error:
        s = ""

    if len(s) != 1:
        retval = proc.wait()
        if retval == 0:
            sys.stderr.write("fork child failed to signal startup\n")
        else:
            # Child exited with an error. Convey the same error to
            # our parent process as a courtesy.
            sys.exit(retval)
    winutils.win32file.CloseHandle(rfd)

    return pid


def _fork_notify_startup(fd):
    if sys.platform == 'win32':
        _fork_notify_startup_windows(fd)
        return

    if fd is not None:
        error, bytes_written = ovs.socket_util.write_fully(fd, "0")
        if error:
            sys.stderr.write("could not write to pipe\n")
            sys.exit(1)
        os.close(fd)


def _fork_notify_startup_windows(fd):
    if fd is not None:
        try:
            # Python 2 requires a string as second parameter, while
            # Python 3 requires a bytes-like object. b"0" fits for both
            # python versions.
            winutils.win32file.WriteFile(fd, b"0", None)
        except winutils.pywintypes.error as e:
            sys.stderr.write("could not write to pipe: %s\n" %
                             os.strerror(e.winerror))
            sys.exit(1)


def _should_restart(status):
    global RESTART_EXIT_CODE

    if sys.platform == 'win32':
        # The exit status is encoded in the high byte of the
        # 16-bit number 'status'.
        exit_status = status >> 8

        if exit_status == RESTART_EXIT_CODE:
            return True
        return False

    if os.WIFEXITED(status) and os.WEXITSTATUS(status) == RESTART_EXIT_CODE:
        return True

    if os.WIFSIGNALED(status):
        for signame in ("SIGABRT", "SIGALRM", "SIGBUS", "SIGFPE", "SIGILL",
                        "SIGPIPE", "SIGSEGV", "SIGXCPU", "SIGXFSZ"):
            if os.WTERMSIG(status) == getattr(signal, signame, None):
                return True
    return False


def _monitor_daemon(daemon_pid):
    # XXX should log daemon's stderr output at startup time
    # XXX should use setproctitle module if available
    last_restart = None
    while True:
        retval, status = _waitpid(daemon_pid, 0)
        if retval < 0:
            sys.stderr.write("waitpid failed\n")
            sys.exit(1)
        elif retval == daemon_pid:
            status_msg = ("pid %d died, %s"
                          % (daemon_pid, ovs.process.status_msg(status)))

            if _should_restart(status):
                if sys.platform != 'win32' and os.WCOREDUMP(status):
                    # Disable further core dumps to save disk space.
                    try:
                        resource.setrlimit(resource.RLIMIT_CORE, (0, 0))
                    except resource.error:
                        vlog.warn("failed to disable core dumps")

                # Throttle restarts to no more than once every 10 seconds.
                if (last_restart is not None and
                    ovs.timeval.msec() < last_restart + 10000):
                    vlog.warn("%s, waiting until 10 seconds since last "
                              "restart" % status_msg)
                    while True:
                        now = ovs.timeval.msec()
                        wakeup = last_restart + 10000
                        if now > wakeup:
                            break
                        sys.stdout.write("sleep %f\n" % (
                            (wakeup - now) / 1000.0))
                        time.sleep((wakeup - now) / 1000.0)
                last_restart = ovs.timeval.msec()

                vlog.err("%s, restarting" % status_msg)
                daemon_pid = _fork_and_wait_for_startup()
                if not daemon_pid:
                    break
            else:
                vlog.info("%s, exiting" % status_msg)
                sys.exit(0)

    # Running in new daemon process.


def _close_standard_fds():
    """Close stdin, stdout, stderr.  If we're started from e.g. an SSH session,
    then this keeps us from holding that session open artificially."""
    null_fd = ovs.socket_util.get_null_fd()
    if null_fd >= 0:
        os.dup2(null_fd, 0)
        os.dup2(null_fd, 1)
        os.dup2(null_fd, 2)


def daemonize_start():
    """If daemonization is configured, then starts daemonization, by forking
    and returning in the child process.  The parent process hangs around until
    the child lets it know either that it completed startup successfully (by
    calling daemon_complete()) or that it failed to start up (by exiting with a
    nonzero exit code)."""

    if _detach:
        if _fork_and_wait_for_startup() > 0:
            # Running in parent process.
            sys.exit(0)

        if sys.platform != 'win32':
            # Running in daemon or monitor process.
            os.setsid()

    if _monitor:
        saved_daemonize_fd = _daemonize_fd
        daemon_pid = _fork_and_wait_for_startup()
        if daemon_pid > 0:
            # Running in monitor process.
            _fork_notify_startup(saved_daemonize_fd)
            if sys.platform != 'win32':
                _close_standard_fds()
            _monitor_daemon(daemon_pid)
        # Running in daemon process

    if _pidfile:
        _make_pidfile()


def daemonize_complete():
    """If daemonization is configured, then this function notifies the parent
    process that the child process has completed startup successfully."""
    _fork_notify_startup(_daemonize_fd)

    if _detach:
        if _chdir:
            os.chdir("/")
        _close_standard_fds()


def usage():
    sys.stdout.write("""
Daemon options:
   --detach                run in background as daemon
   --no-chdir              do not chdir to '/'
   --pidfile[=FILE]        create pidfile (default: %s/%s.pid)
   --overwrite-pidfile     with --pidfile, start even if already running
""" % (ovs.dirs.RUNDIR, ovs.util.PROGRAM_NAME))


def __read_pidfile(pidfile, delete_if_stale):
    if _pidfile_dev is not None:
        try:
            s = os.stat(pidfile)
            if s.st_ino == _pidfile_ino and s.st_dev == _pidfile_dev:
                # It's our own pidfile.  We can't afford to open it,
                # because closing *any* fd for a file that a process
                # has locked also releases all the locks on that file.
                #
                # Fortunately, we know the associated pid anyhow.
                return os.getpid()
        except OSError:
            pass

    try:
        file_handle = open(pidfile, "r+")
    except IOError as e:
        if e.errno == errno.ENOENT and delete_if_stale:
            return 0
        vlog.warn("%s: open: %s" % (pidfile, e.strerror))
        return -e.errno

    # Python fcntl doesn't directly support F_GETLK so we have to just try
    # to lock it.
    try:
        fcntl.lockf(file_handle, fcntl.LOCK_EX | fcntl.LOCK_NB)

        # pidfile exists but wasn't locked by anyone.  Now we have the lock.
        if not delete_if_stale:
            file_handle.close()
            vlog.warn("%s: pid file is stale" % pidfile)
            return -errno.ESRCH

        # Is the file we have locked still named 'pidfile'?
        try:
            raced = False
            s = os.stat(pidfile)
            s2 = os.fstat(file_handle.fileno())
            if s.st_ino != s2.st_ino or s.st_dev != s2.st_dev:
                raced = True
        except IOError:
            raced = True
        if raced:
            vlog.warn("%s: lost race to delete pidfile" % pidfile)
            return -errno.EALREADY

        # We won the right to delete the stale pidfile.
        try:
            os.unlink(pidfile)
        except IOError as e:
            vlog.warn("%s: failed to delete stale pidfile (%s)"
                            % (pidfile, e.strerror))
            return -e.errno
        else:
            vlog.dbg("%s: deleted stale pidfile" % pidfile)
            file_handle.close()
            return 0
    except IOError as e:
        if e.errno not in [errno.EACCES, errno.EAGAIN]:
            vlog.warn("%s: fcntl: %s" % (pidfile, e.strerror))
            return -e.errno

    # Someone else has the pidfile locked.
    try:
        try:
            error = int(file_handle.readline())
        except IOError as e:
            vlog.warn("%s: read: %s" % (pidfile, e.strerror))
            error = -e.errno
        except ValueError:
            vlog.warn("%s does not contain a pid" % pidfile)
            error = -errno.EINVAL

        return error
    finally:
        try:
            file_handle.close()
        except IOError:
            pass


def read_pidfile(pidfile):
    """Opens and reads a PID from 'pidfile'.  Returns the positive PID if
    successful, otherwise a negative errno value."""
    return __read_pidfile(pidfile, False)


def _check_already_running():
    pid = __read_pidfile(_pidfile, True)
    if pid > 0:
        _fatal("%s: already running as pid %d, aborting" % (_pidfile, pid))
    elif pid < 0:
        _fatal("%s: pidfile check failed (%s), aborting"
               % (_pidfile, os.strerror(pid)))


def add_args(parser):
    """Populates 'parser', an ArgumentParser allocated using the argparse
    module, with the command line arguments required by the daemon module."""

    pidfile = make_pidfile_name(None)

    group = parser.add_argument_group(title="Daemon Options")
    group.add_argument("--detach", action="store_true",
            help="Run in background as a daemon.")
    group.add_argument("--no-chdir", action="store_true",
            help="Do not chdir to '/'.")
    group.add_argument("--monitor", action="store_true",
            help="Monitor %s process." % ovs.util.PROGRAM_NAME)
    group.add_argument("--pidfile", nargs="?", const=pidfile,
            help="Create pidfile (default %s)." % pidfile)
    group.add_argument("--overwrite-pidfile", action="store_true",
            help="With --pidfile, start even if already running.")
    if sys.platform == 'win32':
        group.add_argument("--pipe-handle",
                           help=("With --pidfile, start even if "
                                 "already running."))


def handle_args(args):
    """Handles daemon module settings in 'args'.  'args' is an object
    containing values parsed by the parse_args() method of ArgumentParser.  The
    parent ArgumentParser should have been prepared by add_args() before
    calling parse_args()."""

    if sys.platform == 'win32':
        if args.pipe_handle:
            set_detached(args.pipe_handle)

    if args.detach:
        set_detach()

    if args.no_chdir:
        set_no_chdir()

    if args.pidfile:
        set_pidfile(args.pidfile)

    if args.overwrite_pidfile:
        ignore_existing_pidfile()

    if args.monitor:
        set_monitor()



# Copyright (c) 2009, 2010, 2011, 2014, 2016 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import re
import uuid

import ovs.db.parser
import ovs.db.types
import ovs.json
import ovs.jsonrpc
import ovs.ovsuuid
import ovs.poller
import ovs.socket_util
from ovs.db import error


class ConstraintViolation(error.Error):
    def __init__(self, msg, json=None):
        error.Error.__init__(self, msg, json, tag="constraint violation")


def escapeCString(src):
    dst = []
    for c in src:
        if c in "\\\"":
            dst.append("\\" + c)
        elif ord(c) < 32:
            if c == '\n':
                dst.append('\\n')
            elif c == '\r':
                dst.append('\\r')
            elif c == '\a':
                dst.append('\\a')
            elif c == '\b':
                dst.append('\\b')
            elif c == '\f':
                dst.append('\\f')
            elif c == '\t':
                dst.append('\\t')
            elif c == '\v':
                dst.append('\\v')
            else:
                dst.append('\\%03o' % ord(c))
        else:
            dst.append(c)
    return ''.join(dst)


def returnUnchanged(x):
    return x


@functools.total_ordering
class Atom(object):
    __slots__ = ('value', 'type')

    def __init__(self, type_, value=None):
        self.type = type_
        if value is not None:
            self.value = value
        else:
            self.value = type_.default_atom()

    def __eq__(self, other):
        if not isinstance(other, Atom) or self.type != other.type:
            return NotImplemented
        return True if self.value == other.value else False

    def __lt__(self, other):
        if not isinstance(other, Atom) or self.type != other.type:
            return NotImplemented
        return True if self.value < other.value else False

    def __cmp__(self, other):
        if not isinstance(other, Atom) or self.type != other.type:
            return NotImplemented
        elif self.value < other.value:
            return -1
        elif self.value > other.value:
            return 1
        else:
            return 0

    def __hash__(self):
        return hash(self.value)

    @staticmethod
    def default(type_):
        """Returns the default value for the given type_, which must be an
        instance of ovs.db.types.AtomicType.

        The default value for each atomic type is;

          - 0, for integer or real atoms.

          - False, for a boolean atom.

          - "", for a string atom.

          - The all-zeros UUID, for a UUID atom."""
        return Atom(type_)

    def is_default(self):
        return self == self.default(self.type)

    @staticmethod
    def from_json(base, json, symtab=None):
        type_ = base.type
        json = ovs.db.parser.float_to_int(json)
        real_types = [int]
        real_types.extend([float])
        real_types = tuple(real_types)
        if ((type_ == ovs.db.types.IntegerType
                and isinstance(json, int))
            or (type_ == ovs.db.types.RealType
                and isinstance(json, real_types))
            or (type_ == ovs.db.types.BooleanType and isinstance(json, bool))
            or (type_ == ovs.db.types.StringType
                and isinstance(json, str))):
            atom = Atom(type_, json)
        elif type_ == ovs.db.types.UuidType:
            atom = Atom(type_, ovs.ovsuuid.from_json(json, symtab))
        else:
            raise error.Error("expected %s" % type_.to_string(), json)
        atom.check_constraints(base)
        return atom

    @staticmethod
    def from_python(base, value):
        value = ovs.db.parser.float_to_int(value)
        if isinstance(value, base.type.python_types):
            atom = Atom(base.type, value)
        else:
            raise error.Error("expected %s, got %s" % (base.type, type(value)))
        atom.check_constraints(base)
        return atom

    def check_constraints(self, base):
        """Checks whether 'atom' meets the constraints (if any) defined in
        'base' and raises an ovs.db.error.Error if any constraint is violated.

        'base' and 'atom' must have the same type.
        Checking UUID constraints is deferred to transaction commit time, so
        this function does nothing for UUID constraints."""
        assert base.type == self.type
        if base.enum is not None and self not in base.enum:
            raise ConstraintViolation(
                "%s is not one of the allowed values (%s)"
                % (self.to_string(), base.enum.to_string()))
        elif base.type in [ovs.db.types.IntegerType, ovs.db.types.RealType]:
            if ((base.min is None or self.value >= base.min) and
                    (base.max is None or self.value <= base.max)):
                pass
            elif base.min is not None and base.max is not None:
                raise ConstraintViolation(
                    "%s is not in the valid range %.15g to %.15g (inclusive)"
                    % (self.to_string(), base.min, base.max))
            elif base.min is not None:
                raise ConstraintViolation(
                    "%s is less than minimum allowed value %.15g"
                    % (self.to_string(), base.min))
            else:
                raise ConstraintViolation(
                    "%s is greater than maximum allowed value %.15g"
                    % (self.to_string(), base.max))
        elif base.type == ovs.db.types.StringType:
            # XXX The C version validates that the string is valid UTF-8 here.
            # Do we need to do that in Python too?
            s = self.value
            length = len(s)
            if length < base.min_length:
                raise ConstraintViolation(
                    '"%s" length %d is less than minimum allowed length %d'
                    % (s, length, base.min_length))
            elif length > base.max_length:
                raise ConstraintViolation(
                    '"%s" length %d is greater than maximum allowed '
                    'length %d' % (s, length, base.max_length))

    def to_json(self):
        if self.type == ovs.db.types.UuidType:
            return ovs.ovsuuid.to_json(self.value)
        else:
            return self.value

    def cInitAtom(self, var):
        if self.type == ovs.db.types.IntegerType:
            return '.integer = %d' % self.value
        elif self.type == ovs.db.types.RealType:
            return '.real = %.15g' % self.value
        elif self.type == ovs.db.types.BooleanType:
            if self.value:
                return '.boolean = true'
            else:
                return '.boolean = false'
        elif self.type == ovs.db.types.StringType:
            return '.s = %s' % escapeCString(self.value)
        elif self.type == ovs.db.types.UuidType:
            return '.uuid = %s' % ovs.ovsuuid.to_c_assignment(self.value)

    def toEnglish(self, escapeLiteral=returnUnchanged):
        if self.type == ovs.db.types.IntegerType:
            return '%d' % self.value
        elif self.type == ovs.db.types.RealType:
            return '%.15g' % self.value
        elif self.type == ovs.db.types.BooleanType:
            if self.value:
                return 'true'
            else:
                return 'false'
        elif self.type == ovs.db.types.StringType:
            return escapeLiteral(self.value)
        elif self.type == ovs.db.types.UuidType:
            return self.value.value

    __need_quotes_re = re.compile("$|true|false|[^_a-zA-Z]|.*[^-._a-zA-Z]")

    @staticmethod
    def __string_needs_quotes(s):
        return Atom.__need_quotes_re.match(s)

    def to_string(self):
        if self.type == ovs.db.types.IntegerType:
            return '%d' % self.value
        elif self.type == ovs.db.types.RealType:
            return '%.15g' % self.value
        elif self.type == ovs.db.types.BooleanType:
            if self.value:
                return 'true'
            else:
                return 'false'
        elif self.type == ovs.db.types.StringType:
            if Atom.__string_needs_quotes(self.value):
                return ovs.json.to_string(self.value)
            else:
                return self.value
        elif self.type == ovs.db.types.UuidType:
            return str(self.value)

    @staticmethod
    def new(x):
        if isinstance(x, int):
            t = ovs.db.types.IntegerType
        elif isinstance(x, float):
            t = ovs.db.types.RealType
        elif isinstance(x, bool):
            t = ovs.db.types.BooleanType
        elif isinstance(x, str):
            t = ovs.db.types.StringType
        elif isinstance(x, uuid):
            t = ovs.db.types.UuidType
        else:
            raise TypeError
        return Atom(t, x)


@functools.total_ordering
class Datum(object):
    __slots__ = ('type', 'values')

    def __init__(self, type_, values={}):
        self.type = type_
        self.values = values

    def __eq__(self, other):
        if not isinstance(other, Datum):
            return NotImplemented
        return True if self.values == other.values else False

    def __lt__(self, other):
        if not isinstance(other, Datum):
            return NotImplemented
        return True if self.values < other.values else False

    def __cmp__(self, other):
        if not isinstance(other, Datum):
            return NotImplemented
        elif self.values < other.values:
            return -1
        elif self.values > other.values:
            return 1
        else:
            return 0

    __hash__ = None

    def __contains__(self, item):
        return item in self.values

    def copy(self):
        return Datum(self.type, dict(self.values))

    @staticmethod
    def default(type_):
        if type_.n_min == 0:
            values = {}
        elif type_.is_map():
            values = {type_.key.default(): type_.value.default()}
        else:
            values = {type_.key.default(): None}
        return Datum(type_, values)

    def is_default(self):
        return self == Datum.default(self.type)

    def check_constraints(self):
        """Checks that each of the atoms in 'datum' conforms to the constraints
        specified by its 'type' and raises an ovs.db.error.Error.

        This function is not commonly useful because the most ordinary way to
        obtain a datum is ultimately via Datum.from_json() or Atom.from_json(),
        which check constraints themselves."""
        for keyAtom, valueAtom in self.values.items():
            keyAtom.check_constraints(self.type.key)
            if valueAtom is not None:
                valueAtom.check_constraints(self.type.value)

    @staticmethod
    def from_json(type_, json, symtab=None):
        """Parses 'json' as a datum of the type described by 'type'.  If
        successful, returns a new datum.  On failure, raises an
        ovs.db.error.Error.

        Violations of constraints expressed by 'type' are treated as errors.

        If 'symtab' is nonnull, then named UUIDs in 'symtab' are accepted.
        Refer to RFC 7047 for information about this, and for the syntax
        that this function accepts."""
        is_map = type_.is_map()
        if (is_map or
            (isinstance(json, list) and len(json) > 0 and json[0] == "set")):
            if is_map:
                class_ = "map"
            else:
                class_ = "set"

            inner = ovs.db.parser.unwrap_json(json, class_, [list, tuple],
                                              "array")
            n = len(inner)
            if n < type_.n_min or n > type_.n_max:
                raise error.Error("%s must have %d to %d members but %d are "
                                  "present" % (class_, type_.n_min,
                                               type_.n_max, n),
                                  json)

            values = {}
            for element in inner:
                if is_map:
                    key, value = ovs.db.parser.parse_json_pair(element)
                    keyAtom = Atom.from_json(type_.key, key, symtab)
                    valueAtom = Atom.from_json(type_.value, value, symtab)
                else:
                    keyAtom = Atom.from_json(type_.key, element, symtab)
                    valueAtom = None

                if keyAtom in values:
                    if is_map:
                        raise error.Error("map contains duplicate key")
                    else:
                        raise error.Error("set contains duplicate")

                values[keyAtom] = valueAtom

            return Datum(type_, values)
        else:
            keyAtom = Atom.from_json(type_.key, json, symtab)
            return Datum(type_, {keyAtom: None})

    def to_json(self):
        if self.type.is_map():
            return ["map", [[k.to_json(), v.to_json()]
                            for k, v in sorted(self.values.items())]]
        elif len(self.values) == 1:
            key = next(iter(self.values.keys()))
            return key.to_json()
        else:
            return ["set", [k.to_json() for k in sorted(self.values.keys())]]

    def to_string(self):
        head = tail = None
        if self.type.n_max > 1 or len(self.values) == 0:
            if self.type.is_map():
                head = "{"
                tail = "}"
            else:
                head = "["
                tail = "]"

        s = []
        if head:
            s.append(head)

        for i, key in enumerate(sorted(self.values)):
            if i:
                s.append(", ")

            s.append(key.to_string())
            if self.type.is_map():
                s.append("=")
                s.append(self.values[key].to_string())

        if tail:
            s.append(tail)
        return ''.join(s)

    def diff(self, datum):
        if self.type.n_max > 1 or len(self.values) == 0:
            for k, v in datum.values.items():
                if k in self.values and v == self.values[k]:
                    del self.values[k]
                else:
                    self.values[k] = v
        else:
            return datum

        return self

    def as_list(self):
        if self.type.is_map():
            return [[k.value, v.value] for k, v in self.values.items()]
        else:
            return [k.value for k in self.values.keys()]

    def as_dict(self):
        return dict(self.values)

    def as_scalar(self):
        if len(self.values) == 1:
            if self.type.is_map():
                k, v = next(iter(self.values.items()))
                return [k.value, v.value]
            else:
                return next(iter(self.values.keys())).value
        else:
            return None

    def to_python(self, uuid_to_row):
        """Returns this datum's value converted into a natural Python
        representation of this datum's type, according to the following
        rules:

        - If the type has exactly one value and it is not a map (that is,
          self.type.is_scalar() returns True), then the value is:

            * An int or long, for an integer column.

            * An int or long or float, for a real column.

            * A bool, for a boolean column.

            * A str or unicode object, for a string column.

            * A uuid.UUID object, for a UUID column without a ref_table.

            * An object represented the referenced row, for a UUID column with
              a ref_table.  (For the Idl, this object will be an ovs.db.idl.Row
              object.)

          If some error occurs (e.g. the database server's idea of the column
          is different from the IDL's idea), then the default value for the
          scalar type is used (see Atom.default()).

        - Otherwise, if the type is not a map, then the value is a Python list
          whose elements have the types described above.

        - Otherwise, the type is a map, and the value is a Python dict that
          maps from key to value, with key and value types determined as
          described above.

        'uuid_to_row' must be a function that takes a value and an
        ovs.db.types.BaseType and translates UUIDs into row objects."""
        if self.type.is_scalar():
            value = uuid_to_row(self.as_scalar(), self.type.key)
            if value is None:
                return self.type.key.default()
            else:
                return value
        elif self.type.is_map():
            value = {}
            for k, v in self.values.items():
                dk = uuid_to_row(k.value, self.type.key)
                dv = uuid_to_row(v.value, self.type.value)
                if dk is not None and dv is not None:
                    value[dk] = dv
            return value
        else:
            s = set()
            for k in self.values:
                dk = uuid_to_row(k.value, self.type.key)
                if dk is not None:
                    s.add(dk)
            return sorted(s)

    @staticmethod
    def from_python(type_, value, row_to_uuid):
        """Returns a new Datum with the given ovs.db.types.Type 'type_'.  The
        new datum's value is taken from 'value', which must take the form
        described as a valid return value from Datum.to_python() for 'type'.

        Each scalar value within 'value' is initially passed through
        'row_to_uuid', which should convert objects that represent rows (if
        any) into uuid.UUID objects and return other data unchanged.

        Raises ovs.db.error.Error if 'value' is not in an appropriate form for
        'type_'."""
        d = {}
        if isinstance(value, dict):
            for k, v in value.items():
                ka = Atom.from_python(type_.key, row_to_uuid(k))
                va = Atom.from_python(type_.value, row_to_uuid(v))
                d[ka] = va
        elif isinstance(value, (list, set, tuple)):
            for k in value:
                ka = Atom.from_python(type_.key, row_to_uuid(k))
                d[ka] = None
        else:
            ka = Atom.from_python(type_.key, row_to_uuid(value))
            d[ka] = None

        datum = Datum(type_, d)
        datum.check_constraints()
        if not datum.conforms_to_type():
            raise error.Error("%d values when type requires between %d and %d"
                              % (len(d), type_.n_min, type_.n_max))

        return datum

    def __getitem__(self, key):
        if not isinstance(key, Atom):
            key = Atom.new(key)
        if not self.type.is_map():
            raise IndexError
        elif key not in self.values:
            raise KeyError
        else:
            return self.values[key].value

    def get(self, key, default=None):
        if not isinstance(key, Atom):
            key = Atom.new(key)
        if key in self.values:
            return self.values[key].value
        else:
            return default

    def __str__(self):
        return self.to_string()

    def conforms_to_type(self):
        n = len(self.values)
        return self.type.n_min <= n <= self.type.n_max

    def cDeclareDatum(self, name):
        n = len(self.values)
        if n == 0:
            return ["static struct ovsdb_datum %s = { .n = 0 };"]

        s = []
        if self.type.key.type == ovs.db.types.StringType:
            s += ["static struct json %s_key_strings[%d] = {"
                  % (name, n)]
            for key in sorted(self.values):
                s += ['    { .type = JSON_STRING, '
                      '.string = "%s", .count = 2 },'
                      % escapeCString(key.value)]
            s += ["};"]
            s += ["static union ovsdb_atom %s_keys[%d] = {" % (name, n)]
            for i in range(n):
                s += ["    { .s = &%s_key_strings[%d] }," % (name, i)]
            s += ["};"]
        else:
            s = ["static union ovsdb_atom %s_keys[%d] = {" % (name, n)]
            for key in sorted(self.values):
                s += ["    { %s }," % key.cInitAtom(key)]
            s += ["};"]

        if self.type.value:
            if self.type.value.type == ovs.db.types.StringType:
                s += ["static struct json %s_val_strings[%d] = {"
                      % (name, n)]
                for k, v in sorted(self.values):
                    s += ['    { .type = JSON_STRING, '
                          '.string = "%s", .count = 2 },'
                          % escapeCString(v.value)]
                s += ["};"]
                s += ["static union ovsdb_atom %s_values[%d] = {" % (name, n)]
                for i in range(n):
                    s += ["    { .s = &%s_val_strings[%d] }," % (name, i)]
                s += ["};"]
            else:
                s = ["static union ovsdb_atom %s_values[%d] = {" % (name, n)]
                for k, v in sorted(self.values.items()):
                    s += ["    { %s }," % v.cInitAtom(v)]
                s += ["};"]

        s += ["static struct ovsdb_datum %s = {" % name]
        s += ["    .n = %d," % n]
        s += ["    .keys = %s_keys," % name]
        if self.type.value:
            s += ["    .values = %s_values," % name]
        s += ["};"]
        return s



""" Defines the Flow class.
"""


class Section(object):
    """A flow can be seen as composed of different sections, e.g:

     [info] [match] actions=[actions]

    This class represents each of those sections.

    A section is basically a set of Key-Value pairs. Typically, they can be
    expressed as a dictionary, for instance the "match" part of a flow can be
    expressed as:
        {
            "nw_src": "192.168.1.1",
            "nw_dst": "192.168.1.2",
        }
    However, some of them must be expressed as a list which allows for
    duplicated keys. For instance, the "actions" section could be:
        [
            {
                "output": 32
            },
            {
                "output": 33
            }
        ]

    The is_list flag is used to discriminate this.

    Attributes:
        name (str): Name of the section.
        pos (int): Position within the overall flow string.
        string (str): Section string.
        data (list[KeyValue]): Parsed data of the section.
        is_list (bool): Whether the key-values shall be expressed as a list
        (i.e: it allows repeated keys).
    """

    def __init__(self, name, pos, string, data, is_list=False):
        self.name = name
        self.pos = pos
        self.string = string
        self.data = data
        self.is_list = is_list

    def __str__(self):
        return "{} (at {}): {}".format(self.name, self.pos, self.string)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)

    def dict(self):
        return {self.name: self.format_data()}

    def format_data(self):
        """Returns the section's key-values formatted in a dictionary or list
        depending on the value of is_list flag.
        """
        if self.is_list:
            return [{item.key: item.value} for item in self.data]
        else:
            return {item.key: item.value for item in self.data}


class Flow(object):
    """The Flow class is a base class for other types of concrete flows
    (such as OFproto Flows or DPIF Flows).

    A flow is basically comprised of a number of sections.
    For each section named {section_name}, the flow object will have the
    following attributes:
     - {section_name} will return the sections data in a formatted way.
     - {section_name}_kv will return the sections data as a list of KeyValues.

    Args:
        sections (list[Section]): List of sections that comprise the flow
        orig (str): Original flow string.
        id (Any): Optional; identifier that clients can use to uniquely
            identify this flow.
    """

    def __init__(self, sections, orig="", id=None):
        self._sections = sections
        self._orig = orig
        self._id = id
        for section in sections:
            setattr(
                self, section.name, self.section(section.name).format_data()
            )
            setattr(
                self,
                "{}_kv".format(section.name),
                self.section(section.name).data,
            )

    def section(self, name):
        """Return the section by name."""
        return next(
            (sect for sect in self._sections if sect.name == name), None
        )

    @property
    def id(self):
        """Return the Flow ID."""
        return self._id

    @property
    def sections(self):
        """Return the all the sections in a list."""
        return self._sections

    @property
    def orig(self):
        """Return the original flow string."""
        return self._orig

    def dict(self):
        """Returns the Flow information in a dictionary."""
        flow_dict = {"orig": self.orig}
        for section in self.sections:
            flow_dict.update(section.dict())

        return flow_dict



# Copyright (c) 2010, 2012, 2014, 2015 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import ipaddress
import os
import os.path
import random
import socket
import sys

from ovs import dns_resolve
import ovs.fatal_signal
import ovs.poller
import ovs.vlog

try:
    import ssl
except ImportError:
    ssl = None

if sys.platform == 'win32':
    import ovs.winutils as winutils
    import win32file

vlog = ovs.vlog.Vlog("socket_util")


def make_short_name(long_name):
    if long_name is None:
        return None
    long_name = os.path.abspath(long_name)
    long_dirname = os.path.dirname(long_name)
    tmpdir = os.getenv('TMPDIR', '/tmp')
    for x in range(0, 1000):
        link_name = \
            '%s/ovs-un-py-%d-%d' % (tmpdir, random.randint(0, 10000), x)
        try:
            os.symlink(long_dirname, link_name)
            ovs.fatal_signal.add_file_to_unlink(link_name)
            return os.path.join(link_name, os.path.basename(long_name))
        except OSError as e:
            if e.errno != errno.EEXIST:
                break
    raise Exception("Failed to create temporary symlink")


def free_short_name(short_name):
    if short_name is None:
        return
    link_name = os.path.dirname(short_name)
    ovs.fatal_signal.unlink_file_now(link_name)


def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
    """Creates a Unix domain socket in the given 'style' (either
    socket.SOCK_DGRAM or socket.SOCK_STREAM) that is bound to 'bind_path' (if
    'bind_path' is not None) and connected to 'connect_path' (if 'connect_path'
    is not None).  If 'nonblock' is true, the socket is made non-blocking.

    Returns (error, socket): on success 'error' is 0 and 'socket' is a new
    socket object, on failure 'error' is a positive errno value and 'socket' is
    None."""

    try:
        sock = socket.socket(socket.AF_UNIX, style)
    except socket.error as e:
        return get_exception_errno(e), None

    try:
        if nonblock:
            set_nonblocking(sock)
        if bind_path is not None:
            # Delete bind_path but ignore ENOENT.
            try:
                os.unlink(bind_path)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    return e.errno, None

            ovs.fatal_signal.add_file_to_unlink(bind_path)
            sock.bind(bind_path)

            try:
                os.fchmod(sock.fileno(), 0o700)
            except OSError:
                pass
        if connect_path is not None:
            try:
                sock.connect(connect_path)
            except socket.error as e:
                if get_exception_errno(e) != errno.EINPROGRESS:
                    raise
        return 0, sock
    except socket.error as e:
        sock.close()
        if (bind_path is not None and
            os.path.exists(bind_path)):
            ovs.fatal_signal.unlink_file_now(bind_path)
        eno = ovs.socket_util.get_exception_errno(e)
        if (eno == "AF_UNIX path too long" and
            os.uname()[0] == "Linux"):
            short_connect_path = None
            short_bind_path = None
            connect_dirfd = None
            bind_dirfd = None
            # Try workaround using /proc/self/fd
            if connect_path is not None:
                dirname = os.path.dirname(connect_path)
                basename = os.path.basename(connect_path)
                try:
                    connect_dirfd = os.open(dirname,
                                            os.O_DIRECTORY | os.O_RDONLY)
                except OSError as err:
                    return get_exception_errno(err), None
                short_connect_path = "/proc/self/fd/%d/%s" % (connect_dirfd,
                                                              basename)

            if bind_path is not None:
                dirname = os.path.dirname(bind_path)
                basename = os.path.basename(bind_path)
                try:
                    bind_dirfd = os.open(dirname, os.O_DIRECTORY | os.O_RDONLY)
                except OSError as err:
                    return get_exception_errno(err), None
                short_bind_path = "/proc/self/fd/%d/%s" % (bind_dirfd,
                                                           basename)

            try:
                return make_unix_socket(style, nonblock, short_bind_path,
                                        short_connect_path)
            finally:
                if connect_dirfd is not None:
                    os.close(connect_dirfd)
                if bind_dirfd is not None:
                    os.close(bind_dirfd)
        elif (eno == "AF_UNIX path too long"):
            if short:
                return get_exception_errno(e), None
            short_bind_path = None
            try:
                short_bind_path = make_short_name(bind_path)
                short_connect_path = make_short_name(connect_path)
            except:
                free_short_name(short_bind_path)
                return errno.ENAMETOOLONG, None
            try:
                return make_unix_socket(style, nonblock, short_bind_path,
                                        short_connect_path, short=True)
            finally:
                free_short_name(short_bind_path)
                free_short_name(short_connect_path)
        else:
            return get_exception_errno(e), None


def check_connection_completion(sock):
    if sys.platform == "win32":
        p = ovs.poller.SelectPoll()
        event = winutils.get_new_event(None, False, True, None)
        # Receive notification of readiness for writing, of completed
        # connection or multipoint join operation, and of socket closure.
        win32file.WSAEventSelect(sock, event,
                                 win32file.FD_WRITE |
                                 win32file.FD_CONNECT |
                                 win32file.FD_CLOSE)
        p.register(event, ovs.poller.POLLOUT)
    else:
        p = ovs.poller.get_system_poll()
        p.register(sock, ovs.poller.POLLOUT)
    pfds = p.poll(0)
    if len(pfds) == 1:
        revents = pfds[0][1]
        if revents & ovs.poller.POLLERR or revents & ovs.poller.POLLHUP:
            try:
                # The following should raise an exception.
                if ssl and isinstance(sock, ssl.SSLSocket):
                    # SSL wrapped socket does not allow
                    # non-zero optional flag.
                    sock.send("\0".encode())
                else:
                    sock.send("\0".encode(), socket.MSG_DONTWAIT)

                # (Here's where we end up if it didn't.)
                # XXX rate-limit
                vlog.err("poll return POLLERR but send succeeded")
                return errno.EPROTO
            except socket.error as e:
                return get_exception_errno(e)
        else:
            return 0
    else:
        return errno.EAGAIN


def is_valid_ipv4_address(address):
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
    except socket.error:
        return False

    return True


def _inet_parse_active(target, default_port):
    address = target.split(":")
    if len(address) >= 2:
        host_name = ":".join(address[0:-1]).lstrip('[').rstrip(']')
        port = int(address[-1])
    else:
        if default_port:
            port = default_port
        else:
            raise ValueError("%s: port number must be specified" % target)
        host_name = address[0]
    if not host_name:
        raise ValueError("%s: bad peer name format" % target)
    try:
        host_name = str(ipaddress.ip_address(host_name))
    except ValueError:
        host_name = dns_resolve.resolve(host_name)
    if not host_name:
        raise ValueError("%s: bad peer name format" % target)
    return (host_name, port)


def inet_parse_active(target, default_port, raises=True):
    try:
        return _inet_parse_active(target, default_port)
    except ValueError:
        if raises:
            raise
        return ("", default_port)


def inet_create_socket_active(style, address):
    try:
        is_addr_inet = is_valid_ipv4_address(address[0])
        if is_addr_inet:
            sock = socket.socket(socket.AF_INET, style, 0)
            family = socket.AF_INET
        else:
            sock = socket.socket(socket.AF_INET6, style, 0)
            family = socket.AF_INET6
    except socket.error as e:
        return get_exception_errno(e), None

    return family, sock


def inet_connect_active(sock, address, family, dscp):
    try:
        set_nonblocking(sock)
        set_dscp(sock, family, dscp)
        error = sock.connect_ex(address)
        if error not in (0, errno.EINPROGRESS, errno.EWOULDBLOCK):
            sock.close()
            return error
        return 0
    except socket.error as e:
        sock.close()
        return get_exception_errno(e)


def inet_open_active(style, target, default_port, dscp):
    address = inet_parse_active(target, default_port, raises=False)
    family, sock = inet_create_socket_active(style, address)
    if sock is None:
        return family, sock
    error = inet_connect_active(sock, address, family, dscp)
    if error:
        return error, None
    return 0, sock


def get_exception_errno(e):
    """A lot of methods on Python socket objects raise socket.error, but that
    exception is documented as having two completely different forms of
    arguments: either a string or a (errno, string) tuple.  We only want the
    errno."""
    if isinstance(e.args, tuple):
        return e.args[0]
    else:
        return errno.EPROTO


null_fd = -1


def get_null_fd():
    """Returns a readable and writable fd for /dev/null, if successful,
    otherwise a negative errno value.  The caller must not close the returned
    fd (because the same fd will be handed out to subsequent callers)."""
    global null_fd
    if null_fd < 0:
        try:
            # os.devnull ensures compatibility with Windows, returns
            # '/dev/null' for Unix and 'nul' for Windows
            null_fd = os.open(os.devnull, os.O_RDWR)
        except OSError as e:
            vlog.err("could not open %s: %s" % (os.devnull,
                                                os.strerror(e.errno)))
            return -e.errno
    return null_fd


def write_fully(fd, buf):
    """Returns an (error, bytes_written) tuple where 'error' is 0 on success,
    otherwise a positive errno value, and 'bytes_written' is the number of
    bytes that were written before the error occurred.  'error' is 0 if and
    only if 'bytes_written' is len(buf)."""
    bytes_written = 0
    if len(buf) == 0:
        return 0, 0
    if not isinstance(buf, bytes):
        buf = bytes(buf, 'utf-8')
    while True:
        try:
            retval = os.write(fd, buf)
            assert retval >= 0
            if retval == len(buf):
                return 0, bytes_written + len(buf)
            elif retval == 0:
                vlog.warn("write returned 0")
                return errno.EPROTO, bytes_written
            else:
                bytes_written += retval
                buf = buf[:retval]
        except OSError as e:
            return e.errno, bytes_written


def set_nonblocking(sock):
    try:
        sock.setblocking(0)
    except socket.error as e:
        vlog.err("could not set nonblocking mode on socket: %s"
                 % os.strerror(get_exception_errno(e)))


def set_dscp(sock, family, dscp):
    if dscp > 63:
        raise ValueError("Invalid dscp %d" % dscp)

    val = dscp << 2
    if family == socket.AF_INET:
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_TOS, val)
    elif family == socket.AF_INET6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_TCLASS, val)
    else:
        raise ValueError('Invalid family %d' % family)



# Copyright (c) 2016 Cloudbase Solutions Srl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

if sys.platform != 'win32':
    raise Exception("Intended to use only on Windows")
else:
    import ntsecuritycon
    import pywintypes
    import win32con
    import win32event
    import win32file
    import win32pipe
    import win32security
    import winerror


def close_handle(handle, logger=None):
    try:
        win32file.CloseHandle(handle)
        return None
    except pywintypes.error as e:
        if logger is not None:
            logger("failed to close handle: %s" % e.strerror)
        return e.winerror


def windows_create_pipe(sAttrs=-1, nSize=None):
    # Default values if parameters are not passed
    if sAttrs == -1:
        sAttrs = win32security.SECURITY_ATTRIBUTES()
        sAttrs.bInheritHandle = 1
    if nSize is None:
        # If this parameter is zero, the system uses the default buffer size.
        nSize = 0

    try:
        (read_pipe, write_pipe) = win32pipe.CreatePipe(sAttrs, nSize)
    except pywintypes.error:
        raise

    return (read_pipe, write_pipe)


def windows_read_pipe(fd, length):
    try:
        (error, data) = win32file.ReadFile(fd, length)
        return error, data
    except pywintypes.error as e:
        return e.winerror, ""


def create_file(filename, desiredAccess=None, shareMode=None, attributes=-1,
                CreationDisposition=None, flagsAndAttributes=None,
                hTemplateFile=-1):
    # Default values if parameters are not passed
    if desiredAccess is None:
        desiredAccess = win32file.GENERIC_READ | win32file.GENERIC_WRITE
    if shareMode is None:
        shareMode = 0
    if attributes == -1:
        # attributes can be None
        attributes = None
    if CreationDisposition is None:
        CreationDisposition = win32file.OPEN_EXISTING
    if flagsAndAttributes is None:
        flagsAndAttributes = (win32file.FILE_ATTRIBUTE_NORMAL |
                              win32file.FILE_FLAG_OVERLAPPED |
                              win32file.FILE_FLAG_NO_BUFFERING)
    if hTemplateFile == -1:
        hTemplateFile = None

    try:
        npipe = win32file.CreateFile(filename,
                                     desiredAccess,
                                     shareMode,
                                     attributes,
                                     CreationDisposition,
                                     flagsAndAttributes,
                                     hTemplateFile)
    except pywintypes.error:
        raise
    return npipe


def write_file(handle, data, overlapped=None):
    try:
        (errCode, nBytesWritten) = win32file.WriteFile(handle,
                                                       data,
                                                       overlapped)
        # Note: win32file.WriteFile doesn't throw an exception
        # in case it receives ERROR_IO_PENDING.
        return (errCode, nBytesWritten)
    except pywintypes.error as e:
        return (e.winerror, 0)


def read_file(handle, bufsize, overlapped=None):
    try:
        # Note: win32file.ReadFile doesn't throw an exception
        # in case it receives ERROR_IO_PENDING.
        (errCode, read_buffer) = win32file.ReadFile(
            handle, bufsize, overlapped)
        return (errCode, read_buffer)
    except pywintypes.error as e:
        return (e.winerror, "")


def create_named_pipe(pipename, openMode=None, pipeMode=None,
                      nMaxInstances=None, nOutBufferSize=None,
                      nInBufferSize=None, nDefaultTimeOut=None,
                      saAttr=-1):
    # Default values if parameters are not passed
    if openMode is None:
        openMode = win32con.PIPE_ACCESS_DUPLEX | win32con.FILE_FLAG_OVERLAPPED
    if pipeMode is None:
        pipeMode = (win32con.PIPE_TYPE_MESSAGE |
                    win32con.PIPE_READMODE_BYTE |
                    win32con.PIPE_WAIT)
    if nMaxInstances is None:
        nMaxInstances = 64
    if nOutBufferSize is None:
        nOutBufferSize = 65000
    if nInBufferSize is None:
        nInBufferSize = 65000
    if nDefaultTimeOut is None:
        nDefaultTimeOut = 0
    if saAttr == -1:
        # saAttr can be None
        saAttr = win32security.SECURITY_ATTRIBUTES()

        # The identifier authority.
        sia = ntsecuritycon.SECURITY_NT_AUTHORITY

        # Initialize the SID.
        remoteAccessSid = win32security.SID()
        remoteAccessSid.Initialize(
            sia,  # The identifier authority.
            1)  # The number of sub authorities to allocate.
        # Disable access over network.
        remoteAccessSid.SetSubAuthority(
            0,  # The index of the sub authority to set
            ntsecuritycon.SECURITY_NETWORK_RID)

        allowedPsids = []
        # Allow Windows Services to access the Named Pipe.
        allowedPsid_0 = win32security.SID()
        allowedPsid_0.Initialize(
            sia,  # The identifier authority.
            1)  # The number of sub authorities to allocate.
        allowedPsid_0.SetSubAuthority(
            0,  # The index of the sub authority to set
            ntsecuritycon.SECURITY_LOCAL_SYSTEM_RID)
        # Allow Administrators to access the Named Pipe.
        allowedPsid_1 = win32security.SID()
        allowedPsid_1.Initialize(
            sia,  # The identifier authority.
            2)  # The number of sub authorities to allocate.
        allowedPsid_1.SetSubAuthority(
            0,  # The index of the sub authority to set
            ntsecuritycon.SECURITY_BUILTIN_DOMAIN_RID)
        allowedPsid_1.SetSubAuthority(
            1,  # The index of the sub authority to set
            ntsecuritycon.DOMAIN_ALIAS_RID_ADMINS)

        allowedPsids.append(allowedPsid_0)
        allowedPsids.append(allowedPsid_1)

        # Initialize an ACL.
        acl = win32security.ACL()
        acl.Initialize()
        # Add denied ACL.
        acl.AddAccessDeniedAce(win32security.ACL_REVISION,
                               ntsecuritycon.GENERIC_ALL,
                               remoteAccessSid)
        # Add allowed ACLs.
        for allowedPsid in allowedPsids:
            acl.AddAccessAllowedAce(win32security.ACL_REVISION,
                                    ntsecuritycon.GENERIC_ALL,
                                    allowedPsid)

        # Initialize an SD.
        sd = win32security.SECURITY_DESCRIPTOR()
        sd.Initialize()
        # Set DACL.
        sd.SetSecurityDescriptorDacl(True, acl, False)

        saAttr.bInheritHandle = 1
        saAttr.SECURITY_DESCRIPTOR = sd

    try:
        npipe = win32pipe.CreateNamedPipe(pipename,
                                          openMode,
                                          pipeMode,
                                          nMaxInstances,
                                          nOutBufferSize,
                                          nInBufferSize,
                                          nDefaultTimeOut,
                                          saAttr)

        if npipe == win32file.INVALID_HANDLE_VALUE:
            return None

        return npipe
    except pywintypes.error:
        return None


def set_pipe_mode(hPipe, mode=-1, maxCollectionCount=None,
                  collectDataTimeout=None):
    # Default values if parameters are not passed
    if mode == -1:
        mode = win32pipe.PIPE_READMODE_BYTE
    try:
        win32pipe.SetNamedPipeHandleState(
            hPipe, mode, maxCollectionCount, collectDataTimeout)
    except pywintypes.error:
        raise


def connect_named_pipe(pipe_handle, overlapped=None):
    try:
        # If the result of ConnectNamedPipe is ERROR_IO_PENDING or
        # ERROR_PIPE_CONNECTED, then this value is returned.
        # All other error values raise a win32 exception
        error = win32pipe.ConnectNamedPipe(pipe_handle, overlapped)
        return error
    except pywintypes.error as e:
        return e.winerror


def get_pipe_name(name):
    name = name.replace('/', '')
    name = name.replace('\\', '')
    name = "\\\\.\\pipe\\" + name
    return name


def get_overlapped_result(handle, overlapped=None, bWait=False):
    try:
        return win32file.GetOverlappedResult(handle, overlapped, bWait)
    except pywintypes.error:
        raise


def get_new_event(sa=None, bManualReset=True, bInitialState=True,
                  objectName=None):
    return win32event.CreateEvent(sa, bManualReset, bInitialState, objectName)


pipe_disconnected_errors = [winerror.ERROR_PIPE_NOT_CONNECTED,
                            winerror.ERROR_BAD_PIPE,
                            winerror.ERROR_NO_DATA,
                            winerror.ERROR_BROKEN_PIPE]
